@@ -1,13 +1,15 @@
 import os
+import gc
 import sys
 import json
 import copy
 import time
+import types
 import numpy as np
 import pandas as pd
 import logging
-import traceback
 import shutil
+import joblib
 from tabulate import tabulate
 from abc import ABC
 from copy import deepcopy
@@ -35,6 +37,8 @@ from supervised.utils.config import mem
 from supervised.utils.config import LOG_LEVEL
 from supervised.utils.leaderboard_plots import LeaderboardPlots
 from supervised.utils.metric import Metric
+from supervised.utils.metric import UserDefinedEvalMetric
+from supervised.utils.automl_plots import AutoMLPlots
 from supervised.preprocessing.eda import EDA
 from supervised.preprocessing.preprocessing_utils import PreprocessingUtils
 from supervised.tuner.time_controller import TimeController
@@ -45,6 +49,7 @@ from supervised.utils.data_validation import (
     check_greater_than_zero_integer_or_float,
     check_integer,
 )
+from supervised.utils.utils import dump_data, load_data
 
 logger = logging.getLogger(__name__)
 logger.setLevel(LOG_LEVEL)
@@ -235,9 +240,12 @@ class BaseAutoML(BaseEstimator, ABC):
             ldb["metric_value"] += [m.get_final_loss()]
             ldb["train_time"] += [np.round(m.get_train_time(), 2)]
             if self._max_single_prediction_time is not None:
-                ldb["single_prediction_time"] += [
-                    np.round(m._single_prediction_time, 4)
-                ]
+                if m._single_prediction_time is not None:
+                    ldb["single_prediction_time"] += [
+                        np.round(m._single_prediction_time, 4)
+                    ]
+                else:
+                    ldb["single_prediction_time"] += [None]
 
         ldb = pd.DataFrame(ldb)
         # need to add argument for sorting
@@ -356,11 +364,12 @@ class BaseAutoML(BaseEstimator, ABC):
         )
         mf.train(results_path, model_subpath)
 
+        # keep info about the model
+        self.keep_model(mf, model_subpath)
+
         # save the model
         mf.save(results_path, model_subpath)
 
-        # and keep info about the model
-        self.keep_model(mf, model_subpath)
         return True
 
     def verbose_print(self, msg):
@@ -383,8 +392,8 @@ class BaseAutoML(BaseEstimator, ABC):
             )
             oofs, target, sample_weight = self.ensemble.get_oof_matrix(self._models)
             self.ensemble.fit(oofs, target, sample_weight)
-            self.ensemble.save(self._results_path, ensemble_subpath)
             self.keep_model(self.ensemble, ensemble_subpath)
+            self.ensemble.save(self._results_path, ensemble_subpath)
             return True
         return False
 
@@ -475,18 +484,18 @@ class BaseAutoML(BaseEstimator, ABC):
 
         self._perform_model_stacking()
 
-        X_stacked_path = os.path.join(self._results_path, "X_stacked.parquet")
+        X_stacked_path = os.path.join(self._results_path, "X_stacked.data")
         if os.path.exists(X_stacked_path):
             return
 
-        X = pd.read_parquet(self._X_path)
+        X = load_data(self._X_path)
         org_columns = X.columns.tolist()
         X_stacked = self.get_stacked_data(X)
         new_columns = X_stacked.columns.tolist()
         added_columns = [c for c in new_columns if c not in org_columns]
 
         # save stacked train data
-        X_stacked.to_parquet(X_stacked_path, index=False)
+        dump_data(X_stacked_path, X_stacked)
 
         """
         # resue old params
@@ -518,7 +527,7 @@ class BaseAutoML(BaseEstimator, ABC):
             self.train_model(params)
         """
 
-    def _save_data(self, X, y, sample_weight=None):
+    def _save_data(self, X, y, sample_weight=None, cv=None):
         # save information about original data
         self._save_data_info(X, y, sample_weight)
 
@@ -528,23 +537,23 @@ class BaseAutoML(BaseEstimator, ABC):
         self._handle_drastic_imbalance(X, y, sample_weight)
 
         # prepare path for saving files
-        self._X_path = os.path.join(self._results_path, "X.parquet")
-        self._y_path = os.path.join(self._results_path, "y.parquet")
+        self._X_path = os.path.join(self._results_path, "X.data")
+        self._y_path = os.path.join(self._results_path, "y.data")
         self._sample_weight_path = None
         if sample_weight is not None:
             self._sample_weight_path = os.path.join(
-                self._results_path, "sample_weight.parquet"
+                self._results_path, "sample_weight.data"
             )
-            pd.DataFrame({"sample_weight": sample_weight}).to_parquet(
-                self._sample_weight_path, index=False
+            dump_data(
+                self._sample_weight_path, pd.DataFrame({"sample_weight": sample_weight})
             )
 
-        X.to_parquet(self._X_path, index=False)
+        dump_data(self._X_path, X)
 
         if self._ml_task == MULTICLASS_CLASSIFICATION:
             y = y.astype(str)
 
-        pd.DataFrame({"target": y}).to_parquet(self._y_path, index=False)
+        dump_data(self._y_path, pd.DataFrame({"target": y}))
 
         # set paths in validation parameters
         self._validation_strategy["X_path"] = self._X_path
@@ -553,10 +562,16 @@ class BaseAutoML(BaseEstimator, ABC):
         if sample_weight is not None:
             self._validation_strategy["sample_weight_path"] = self._sample_weight_path
 
+        if cv is not None:
+            self._validation_strategy["cv_path"] = os.path.join(
+                self._results_path, "cv.data"
+            )
+            joblib.dump(cv, self._validation_strategy["cv_path"])
+
         if self._max_single_prediction_time is not None:
             self._one_sample = X.iloc[:1].copy(deep=True)
 
-        self._drop_data_variables(X)
+        # self._drop_data_variables(X)
 
     def _handle_drastic_imbalance(self, X, y, sample_weight=None):
         if self._ml_task == REGRESSION:
@@ -626,7 +641,7 @@ class BaseAutoML(BaseEstimator, ABC):
 
     def _load_data_variables(self, X_train):
         if X_train.shape[1] == 0:
-            X = pd.read_parquet(self._X_path)
+            X = load_data(self._X_path)
             for c in X.columns:
                 X_train.insert(loc=X_train.shape[1], column=c, value=X[c])
 
@@ -821,7 +836,10 @@ class BaseAutoML(BaseEstimator, ABC):
             del self._validation_strategy["train_ratio"]
             self._validation_strategy["k_folds"] = k_folds
             self.tuner._validation_strategy = self._validation_strategy
-            shutil.rmtree(self._model_subpaths[0], ignore_errors=True)
+            shutil.rmtree(
+                os.path.join(self._results_path, self._model_subpaths[0]),
+                ignore_errors=True,
+            )
             del self._models[0]
             del self._model_subpaths[0]
             del self.tuner._unique_params_keys[0]
@@ -871,7 +889,7 @@ class BaseAutoML(BaseEstimator, ABC):
             ):
                 self._time_ctrl._steps.remove("boost_on_errors")
 
-    def _fit(self, X, y, sample_weight=None):
+    def _fit(self, X, y, sample_weight=None, cv=None):
         """Fits the AutoML model with data"""
         if self._fit_level == "finished":
             print(
@@ -967,7 +985,7 @@ class BaseAutoML(BaseEstimator, ABC):
                 EDA.compute(X, y, os.path.join(self._results_path, "EDA"))
 
             # Save data
-            self._save_data(X.copy(deep=False), y, sample_weight)
+            self._save_data(X.copy(deep=False), y, sample_weight, cv)
 
             tuner = MljarTuner(
                 self._get_tuner_params(
@@ -1089,6 +1107,8 @@ class BaseAutoML(BaseEstimator, ABC):
                             params.get("name") + " not trained. " + str(e)
                         )
                     except Exception as e:
+                        import traceback
+
                         self._update_errors_report(
                             params.get("name"), str(e) + "\n" + traceback.format_exc()
                         )
@@ -1109,9 +1129,9 @@ class BaseAutoML(BaseEstimator, ABC):
 
         except Exception as e:
             raise e
-        finally:
-            if self._X_path is not None:
-                self._load_data_variables(X)
+        # finally:
+        #    if self._X_path is not None:
+        #        self._load_data_variables(X)
 
         return self
 
@@ -1225,6 +1245,9 @@ class BaseAutoML(BaseEstimator, ABC):
                 fout.write(tabulate(ldb.values, ldb.columns, tablefmt="pipe"))
                 LeaderboardPlots.compute(ldb, self._results_path, fout)
 
+                if self._fit_level == "finished":
+                    AutoMLPlots.add(self._results_path, self._models, fout)
+
     def _base_predict(self, X, model=None):
 
         if model is None:
@@ -1464,6 +1487,9 @@ class BaseAutoML(BaseEstimator, ABC):
         """ Gets the current stack_models"""
         self._validate_stack_models()
         if self.stack_models == "auto":
+            val = self._get_validation_strategy()
+            if val.get("validation_type", "") == "custom":
+                return False
             return True if self.mode in ["Compete", "Optuna"] else False
         else:
             return deepcopy(self.stack_models)
@@ -1471,6 +1497,10 @@ class BaseAutoML(BaseEstimator, ABC):
     def _get_eval_metric(self):
         """ Gets the current eval_metric"""
         self._validate_eval_metric()
+        if isinstance(self.eval_metric, types.FunctionType):
+            UserDefinedEvalMetric().set_metric(self.eval_metric)
+            return "user_defined_metric"
+
         if self.eval_metric == "auto":
             if self._get_ml_task() == BINARY_CLASSIFICATION:
                 return "logloss"
@@ -1618,6 +1648,9 @@ class BaseAutoML(BaseEstimator, ABC):
         """ Gets the current boost_on_errors"""
         self._validate_boost_on_errors()
         if self.boost_on_errors == "auto":
+            val = self._get_validation_strategy()
+            if val.get("validation_type", "") == "custom":
+                return False
             if self._get_mode() == "Explain":
                 return False
             if self._get_mode() == "Perform":
@@ -1625,7 +1658,7 @@ class BaseAutoML(BaseEstimator, ABC):
             if self._get_mode() == "Compete":
                 return True
             if self._get_mode() == "Optuna":
-                return True
+                return False
         else:
             return deepcopy(self.boost_on_errors)
 
@@ -1771,7 +1804,9 @@ class BaseAutoML(BaseEstimator, ABC):
 
     def _validate_eval_metric(self):
         """ Validates eval_metric parameter"""
-        # `stack_models` defaults to "auto". If not "auto", check if is valid bool
+        if isinstance(self.eval_metric, types.FunctionType):
+            return
+
         if isinstance(self.eval_metric, str) and self.eval_metric == "auto":
             return
 
@@ -1780,7 +1815,7 @@ class BaseAutoML(BaseEstimator, ABC):
             "auc",
             "f1",
             "average_precision",
-            "accuracy"
+            "accuracy",
         ]:
             raise ValueError(
                 f"Metric {self.eval_metric} is not allowed in ML task: {self._get_ml_task()}. \
@@ -1850,6 +1885,8 @@ class BaseAutoML(BaseEstimator, ABC):
         """ Validates golden_features parameter"""
         if isinstance(self.golden_features, str) and self.golden_features == "auto":
             return
+        if isinstance(self.golden_features, int):
+            return
         check_bool(self.golden_features, "golden_features")
 
     def _validate_features_selection(self):
@@ -1965,57 +2002,7 @@ class BaseAutoML(BaseEstimator, ABC):
 
         self._ml_task = json_data.get("ml_task")
 
-    def _md_to_html(self, md_fname, page_type):
-        import markdown
-
-        if not os.path.exists(md_fname):
-            return None
-        content = ""
-        with open(md_fname) as fin:
-            content = fin.read()
-
-        content = content.replace("README.md", "README.html")
-        content_html = markdown.markdown(
-            content, extensions=["markdown.extensions.tables"]
-        )
-        content_html = content_html.replace("<img ", '<img style="width:80%" ')
-        content_html = content_html.replace("<table>", '<table class="styled-table">')
-        content_html = content_html.replace("<tr>", '<tr style="text-align: right;">')
-
-        styles = '<link rel="stylesheet" href="style.css">\n\n'
-        if page_type == "sub":
-            styles = '<link rel="stylesheet" href="../style.css">\n\n'
-        beginning = styles
-
-        if page_type == "main":
-            beginning += """<img src="https://raw.githubusercontent.com/mljar/visual-identity/main/media/mljar_AutomatedML.png" style="height:128px; margin-left: auto;
-margin-right: auto;display: block;"/>\n\n"""
-            if os.path.exists(os.path.join(self._results_path, "EDA")):
-                beginning += '<a href="EDA/README.html">Automatic Exploratory Data Analysis Report</a>'
-
-        content_html = beginning + content_html
-
-        html_fname = md_fname.replace("README.md", "README.html")
-        with open(html_fname, "w") as fout:
-            fout.write(content_html)
-
-        return html_fname
-
-    def _report(self, width=900, height=1200):
-
-        from IPython.display import IFrame
-
-        main_readme_html = os.path.join(self._results_path, "README.html")
-        if not os.path.exists(main_readme_html) or 1:
-            fname = os.path.join(self._results_path, "README.md")
-            main_readme_html = self._md_to_html(fname, "main")
-            for f in os.listdir(self._results_path):
-                fname = os.path.join(self._results_path, f, "README.md")
-                if os.path.exists(fname):
-                    self._md_to_html(fname, "sub")
-            with open(os.path.join(self._results_path, "style.css"), "w") as fout:
-                fout.write(
-                    """
+    report_style = """
 .styled-table {
     border-collapse: collapse;
     font-size: 0.9em;
@@ -2031,7 +2018,7 @@ margin-right: auto;display: block;"/>\n\n"""
 
 .styled-table tr:hover {background-color: #e0ecf5;}
 
-.styled-table th {
+.styled-table thead {
     padding-top: 6px;
     padding-bottom: 6px;
     text-align: left;
@@ -2041,6 +2028,7 @@ margin-right: auto;display: block;"/>\n\n"""
 
 body {
     font-family: Arial;
+    font-size: 1.0em;
     background-color: rgba(236, 243, 249, 0.15);
 }
 
@@ -2067,12 +2055,157 @@ h3 {
     padding-bottom: 5px;
     margin-bottom: 0px;
 }
+a {
+    font-weight: bold;
+    color: #004666;
+}
+
+a:hover {
+    cursor: pointer;
+    color: #0099CC;
+}
+
 
 """
+
+    def _md_to_html(self, md_fname, page_type, dir_path, me=None):
+        import markdown
+        import base64
+
+        if not os.path.exists(md_fname):
+            return None
+        content = ""
+        with open(md_fname) as fin:
+            content = fin.read()
+
+        content = content.replace("README.md", "README.html")
+        content_html = markdown.markdown(
+            content, extensions=["markdown.extensions.tables"]
+        )
+        content_html = content_html.replace("<img ", '<img style="width:750px" ')
+        content_html = content_html.replace("<table>", '<table class="styled-table">')
+        content_html = content_html.replace("<tr>", '<tr style="text-align: right;">')
+
+        # replace png figures to base64
+        for f in os.listdir(dir_path):
+            if ".png" in f:
+                encoded_string = ""
+                with open(os.path.join(dir_path, f), "rb") as image_file:
+                    encoded_string = base64.b64encode(image_file.read())
+                    encoded_string = encoded_string.decode("utf-8")
+                encoded_figure = f"data:image/png;base64, {encoded_string}"
+                content_html = content_html.replace(f, encoded_figure)
+
+        # insert svg figures
+        for f in os.listdir(dir_path):
+            if ".svg" in f:
+                with open(os.path.join(dir_path, f), "rb") as image_file:
+                    svg_plot = image_file.read()
+                    svg_plot = svg_plot.decode("utf-8")
+
+                arr = content_html.split("\n")
+                new_content = []
+                for i in arr:
+                    if f in i:
+                        new_content += [f"<p>{svg_plot}</p>"]
+                    else:
+                        new_content += [i]
+                content_html = "\n".join(new_content)
+
+        # change links
+        if page_type == "main":
+            for f in os.listdir(dir_path):
+                if os.path.exists(os.path.join(dir_path, f, "README.md")):
+                    old = f'href="{f}/README.html"'
+                    new = f"onclick=\"toggleShow('{f}');toggleShow('main')\" "
+                    content_html = content_html.replace(old, new)
+
+        # other links
+        if me is not None:
+            old = 'href="../README.html"'
+            new = f"onclick=\"toggleShow('{me}');toggleShow('main')\" "
+            content_html = content_html.replace(old, new)
+
+        beginning = ""
+
+        if page_type == "main":
+            beginning += """<img src="https://raw.githubusercontent.com/mljar/visual-identity/main/media/mljar_AutomatedML.png" style="height:128px; margin-left: auto;
+margin-right: auto;display: block;"/>\n\n"""
+            if os.path.exists(os.path.join(self._results_path, "EDA")):
+                beginning += "<a onclick=\"toggleShow('EDA');toggleShow('main')\" >Automatic Exploratory Data Analysis Report</a>"
+            if os.path.exists(os.path.join(self._results_path, "optuna/README.md")):
+                beginning += "<h2><a onclick=\"toggleShow('optuna');toggleShow('main')\" >&#187; Optuna Params Tuning Report</a></h2>"
+
+        content_html = beginning + content_html
+
+        return content_html
+
+    def _show_report(self, main_readme_html, width=900, height=1200):
+        from IPython.display import HTML, IFrame
+
+        if os.environ.get("KAGGLE_KERNEL_RUN_TYPE") is None:
+            with open(main_readme_html) as fin:
+                return HTML(fin.read())
+        else:
+            return IFrame(main_readme_html, width=width, height=height)
+
+    def _report(self, width=900, height=1200):
+
+        self._results_path = self._get_results_path()
+        main_readme_html = os.path.join(self._results_path, "README.html")
+
+        if os.path.exists(main_readme_html):
+            return self._show_report(main_readme_html, width, height)
+
+        body = ""
+        fname = os.path.join(self._results_path, "README.md")
+        body += (
+            '<div id="main">\n'
+            + self._md_to_html(fname, "main", self._results_path)
+            + "\n\n</div>\n\n"
+        )
+
+        for f in os.listdir(self._results_path):
+            fname = os.path.join(self._results_path, f, "README.md")
+            if os.path.exists(fname):
+                body += (
+                    f'<div id="{f}" style="display: none">\n'
+                    + self._md_to_html(
+                        fname, "sub", os.path.join(self._results_path, f), f
+                    )
+                    + "\n\n</div>\n\n"
                 )
 
-        if main_readme_html is not None:
-            return IFrame(main_readme_html, width, height)
+        body += """
+    <script>
+        function toggleShow(elementId) {
+            var x = document.getElementById(elementId);
+            if (x.style.display === "none") {
+                x.style.display = "block";
+            } else {
+                x.style.display = "none";
+            }
+        }
+    </script>
+        """
+
+        report_content = f"""
+<!DOCTYPE html>
+<html>
+<head>
+    <style>
+    {self.report_style}
+    </style>
+</head>
+<body>
+    {body}
+</body>
+</html>
+"""
+        with open(main_readme_html, "w") as fout:
+            fout.write(report_content)
+
+        return self._show_report(main_readme_html, width, height)
 
     def _need_retrain(self, X, y, sample_weight, decrease):
 
