@@ -20,7 +20,7 @@ from supervised.algorithms.registry import (
 from supervised.algorithms.xgboost import xgboost_eval_metric
 from supervised.algorithms.lightgbm import lightgbm_eval_metric
 from supervised.algorithms.catboost import catboost_eval_metric
-
+from supervised.utils.utils import dump_data
 import logging
 from supervised.utils.config import LOG_LEVEL
 
@@ -262,8 +262,9 @@ class MljarTuner:
             # didnt find anything matching the step, return empty array
             return []
         except Exception as e:
-            # import traceback
-            # print(traceback.format_exc())
+            import traceback
+
+            print(str(e), traceback.format_exc())
             return []
 
     def get_params_stack_models(self, stacked_models):
@@ -302,7 +303,7 @@ class MljarTuner:
 
             params["validation_strategy"]["X_path"] = params["validation_strategy"][
                 "X_path"
-            ].replace("X.parquet", "X_stacked.parquet")
+            ].replace("X.data", "X_stacked.data")
 
             params["name"] = params["name"] + "_Stacked"
             params["is_stacked"] = True
@@ -744,6 +745,15 @@ class MljarTuner:
                 "results_path": results_path,
                 "ml_task": self._ml_task,
             }
+            if (
+                self._golden_features is not None
+                and not isinstance(self._golden_features, bool)
+                and isinstance(self._golden_features, int)
+            ):
+                params["preprocessing"]["golden_features"][
+                    "features_count"
+                ] = self._golden_features
+
             params["name"] += "_GoldenFeatures"
             params["status"] = "initialized"
             params["final_loss"] = None
@@ -1048,10 +1058,10 @@ class MljarTuner:
         df_preds = df_preds.sort_values(by="lp", ascending=True)
 
         sample_weight_path = os.path.join(
-            results_path, best_model.get_name() + "_sample_weight.parquet"
+            results_path, best_model.get_name() + "_sample_weight.data"
         )
-        pd.DataFrame({"sample_weight": df_preds["order"]}).to_parquet(
-            sample_weight_path, index=False
+        dump_data(
+            sample_weight_path, pd.DataFrame({"sample_weight": df_preds["order"]})
         )
 
         generated_params = []
